use std::{fmt, str::FromStr};

use diesel::{
    deserialize::{FromSql, FromSqlRow},
    expression::AsExpression,
    prelude::*,
    serialize::ToSql,
    sql_types::Text,
    sqlite::{Sqlite, SqliteValue},
};
use serde::{Deserialize, Serialize};

use crate::db::{schema::*, DbPool, DbResult};

#[derive(Serialize, Deserialize, Clone, Queryable, Selectable)]
#[diesel(table_name = devices)]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct Device {
    pub id: i64,
    pub device_id: String,
    pub user_id: i64,
    pub caption: String,
    pub type_: DeviceType,
}

#[derive(Deserialize, Insertable)]
#[diesel(table_name = devices)]
#[diesel(check_for_backend(diesel::sqlite::Sqlite))]
pub struct NewDevice {
    pub device_id: String,
    pub user_id: i64,
    pub caption: String,
    pub type_: DeviceType,
}

#[derive(Serialize, Deserialize, FromSqlRow, Debug, AsExpression, Clone)]
#[diesel(sql_type = Text)]
#[serde(rename_all = "lowercase")]
pub enum DeviceType {
    Desktop,
    Laptop,
    Mobile,
    Server,
    Other,
}

impl Device {
    pub fn for_user(pool: &DbPool, user_id: i64) -> DbResult<Vec<Self>> {
        Ok(devices::dsl::devices
            .select(Self::as_select())
            .filter(devices::user_id.eq(user_id))
            .get_results(&mut pool.get()?)?)
    }

    pub fn device_id_to_id(
        conn: &mut SqliteConnection,
        user_id: i64,
        device_id: &str,
    ) -> diesel::QueryResult<i64> {
        devices::table
            .select(devices::id)
            .filter(
                devices::user_id
                    .eq(user_id)
                    .and(devices::device_id.eq(device_id)),
            )
            .get_result(conn)
    }

    pub fn by_device_id(
        conn: &mut SqliteConnection,
        user_id: i64,
        device_id: &str,
    ) -> diesel::QueryResult<Self> {
        devices::dsl::devices
            .select(Self::as_select())
            .filter(
                devices::user_id
                    .eq(user_id)
                    .and(devices::device_id.eq(device_id)),
            )
            .get_result(conn)
    }

    pub fn update(&self, pool: &DbPool) -> DbResult<()> {
        Ok(diesel::update(
            devices::table.filter(
                devices::user_id
                    .eq(self.user_id)
                    .and(devices::device_id.eq(&self.device_id)),
            ),
        )
        .set((
            devices::caption.eq(&self.caption),
            devices::type_.eq(&self.type_),
        ))
        .execute(&mut pool.get()?)
        .map(|_| ())?)
    }
}

impl NewDevice {
    pub fn new(user_id: i64, device_id: String, caption: String, type_: DeviceType) -> Self {
        Self {
            device_id,
            user_id,
            caption,
            type_,
        }
    }

    pub fn insert(self, pool: &DbPool) -> DbResult<Device> {
        Ok(diesel::insert_into(devices::table)
            .values(&self)
            .returning(Device::as_returning())
            .get_result(&mut pool.get()?)?)
    }
}

impl fmt::Display for DeviceType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Desktop => "desktop",
                Self::Laptop => "laptop",
                Self::Mobile => "mobile",
                Self::Server => "server",
                Self::Other => "other",
            }
        )
    }
}

#[derive(Debug)]
pub struct DeviceTypeParseErr(String);

impl fmt::Display for DeviceTypeParseErr {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "invalid device type '{}'", self.0)
    }
}

impl std::error::Error for DeviceTypeParseErr {}

impl FromStr for DeviceType {
    type Err = DeviceTypeParseErr;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "desktop" => Ok(Self::Desktop),
            "laptop" => Ok(Self::Laptop),
            "mobile" => Ok(Self::Mobile),
            "server" => Ok(Self::Server),
            "other" => Ok(Self::Other),
            _ => Err(DeviceTypeParseErr(s.to_string())),
        }
    }
}

impl FromSql<Text, Sqlite> for DeviceType {
    fn from_sql(bytes: SqliteValue) -> diesel::deserialize::Result<Self> {
        let s = <String as FromSql<Text, Sqlite>>::from_sql(bytes)?;

        Ok(s.as_str().parse()?)
    }
}

impl ToSql<Text, Sqlite> for DeviceType {
    fn to_sql<'b>(
        &'b self,
        out: &mut diesel::serialize::Output<'b, '_, Sqlite>,
    ) -> diesel::serialize::Result {
        out.set_value(self.to_string());

        Ok(diesel::serialize::IsNull::No)
    }
}
