use diesel::prelude::*;

use super::SqliteRepository;
use crate::{
    db::{self, schema::*},
    gpodder,
};

impl From<db::DeviceType> for gpodder::DeviceType {
    fn from(value: db::DeviceType) -> Self {
        match value {
            db::DeviceType::Desktop => Self::Desktop,
            db::DeviceType::Laptop => Self::Laptop,
            db::DeviceType::Mobile => Self::Mobile,
            db::DeviceType::Server => Self::Server,
            db::DeviceType::Other => Self::Other,
        }
    }
}

impl From<gpodder::DeviceType> for db::DeviceType {
    fn from(value: gpodder::DeviceType) -> Self {
        match value {
            gpodder::DeviceType::Desktop => Self::Desktop,
            gpodder::DeviceType::Laptop => Self::Laptop,
            gpodder::DeviceType::Mobile => Self::Mobile,
            gpodder::DeviceType::Server => Self::Server,
            gpodder::DeviceType::Other => Self::Other,
        }
    }
}

impl gpodder::DeviceRepository for SqliteRepository {
    fn devices_for_user(
        &self,
        user: &gpodder::User,
    ) -> Result<Vec<gpodder::Device>, gpodder::AuthErr> {
        Ok(devices::table
            .select(db::Device::as_select())
            .filter(devices::user_id.eq(user.id))
            .get_results(&mut self.pool.get()?)?
            .into_iter()
            .map(|d| gpodder::Device {
                id: d.device_id,
                caption: d.caption,
                r#type: d.type_.into(),
                // TODO implement subscription count
                subscriptions: 0,
            })
            .collect())
    }

    fn update_device_info(
        &self,
        user: &gpodder::User,
        device_id: &str,
        patch: gpodder::DevicePatch,
    ) -> Result<(), gpodder::AuthErr> {
        if let Some(mut device) = devices::table
            .select(db::Device::as_select())
            .filter(
                devices::user_id
                    .eq(user.id)
                    .and(devices::device_id.eq(device_id)),
            )
            .get_result(&mut self.pool.get()?)
            .optional()?
        {
            if let Some(caption) = patch.caption {
                device.caption = caption;
            }

            if let Some(type_) = patch.r#type {
                device.type_ = type_.into();
            }

            diesel::update(devices::table.filter(devices::id.eq(device.id)))
                .set((
                    devices::caption.eq(&device.caption),
                    devices::type_.eq(&device.type_),
                ))
                .execute(&mut self.pool.get()?)?;
        } else {
            let device = db::NewDevice {
                device_id: device_id.to_string(),
                user_id: user.id,
                caption: patch.caption.unwrap_or(String::new()),
                type_: patch.r#type.unwrap_or(gpodder::DeviceType::Other).into(),
            };

            diesel::insert_into(devices::table)
                .values(device)
                .execute(&mut self.pool.get()?)?;
        }

        Ok(())
    }
}
