use std::ops::Deref;

use serde::{
    de::{value::StrDeserializer, Visitor},
    Deserialize,
};

#[derive(Deserialize, Debug, PartialEq, Eq)]
#[serde(rename_all = "lowercase")]
pub enum Format {
    Json,
    // OPML,
    // Plaintext,
}

#[derive(Debug)]
pub struct StringWithFormat {
    pub s: String,
    pub format: Format,
}

impl Deref for StringWithFormat {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.s
    }
}

impl<'de> Deserialize<'de> for StringWithFormat {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        struct StrVisitor;

        impl<'de> Visitor<'de> for StrVisitor {
            type Value = StringWithFormat;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str(
                    "`text.ext` format, with `ext` being one of `json`, `opml` or `plaintext`",
                )
            }

            fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
            where
                E: serde::de::Error,
            {
                if let Some((text, ext)) = v.rsplit_once('.') {
                    let format = Format::deserialize(StrDeserializer::new(ext))?;

                    Ok(StringWithFormat {
                        s: text.to_string(),
                        format,
                    })
                } else {
                    Err(E::custom(format!("invalid format '{}'", v)))
                }
            }
        }

        deserializer.deserialize_str(StrVisitor)
    }
}
